#ifndef __MATH3D_INC
#define __MATH3D_INC

//
// General (3d) math lib with some usefull classes
// Copyright (c) 1999 by B.Pickhardt (pickhardt@debitel.net).
// aka hackCraft
//



//--[Include Files]----------------->

#ifdef __win32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#endif

// ^- ps. don't ask why...



#include <stdio.h>
#include <math.h>

extern FILE                 *errlog; // error log file


//--[Machine Specific]-------------->


typedef  float           FLOAT;

typedef  long            INTEG;

typedef  unsigned char   BOOLEAN;

typedef  unsigned long   FLAGS;

typedef  unsigned char   BYTE;

typedef  unsigned short  WORD;

typedef  unsigned long   DWORD;


//--[Non Machine Specific]---------->


typedef  FLOAT           ANGLE;

typedef  FLOAT           SCALAR;

typedef  SCALAR          VECTOR[4]; // homogenous /16 byte aligned /plane /rgba

typedef  SCALAR          MATRIX[4*4];


// NOTE: the 4th element of VECTOR[] is initialiased by the Vector class
// but that element isn't used at all except in the Plane class.
// It's up to the programmer for what the 4th element should be used
// in the Vector class - if at all - except in the Plane class of course.


//--[Array Accessors]--------------->

#define  X 0
#define  Y 1
#define  Z 2
#define  W 3
#define  D 3

// better try not to use that letters for variables etc.
// ...but who would use that short names anyway?


//--[classes]----------------------->

class cVector;
class cMatrix;
class cPlane;
class cSpline;
class cRay;


class cVector {
  public:
  VECTOR            mV; // we only realy care about X, Y and Z
  public:
                    cVector();
  BOOLEAN           Equal(cVector v);
  BOOLEAN           Unequal(cVector v);
  void              Set(SCALAR x, SCALAR y, SCALAR z, SCALAR w);
  void              Set(SCALAR x, SCALAR y, SCALAR z);
  void              Set(cVector p0, cVector p1);
  INTEG             Perspective(SCALAR xzoom, SCALAR yzoom);
  void              Polar();
  void              Normalise();
  SCALAR            Len();
  SCALAR            Dot(cVector v);
  void              Crs(cVector v);
  void              Add(cVector v);
  void              Sub(cVector v);
  void              Mul(SCALAR s);
  void              Div(SCALAR s);
  void              Neg();
  ANGLE             XZangle();
  ANGLE             XYangle();
  ANGLE             YZangle();
  INTEG             Nearest(cVector *vList, INTEG n);
  INTEG             Furthest(cVector *vList, INTEG n);
  void              CenterOf(cVector *vList, INTEG n);
  INTEG             FindInList(cVector *vList, INTEG n);
  BOOLEAN           InPolygon(INTEG n, cVector *vtx);
  cVector           GetUV(INTEG n, cVector *vtx, cVector *tex);
  void              Print(FILE *f);
};


class cPlane {
  public:
  cVector           mV; // now we use the D (W) beside the X, Y and Z, too!
  public:
  void              Set(cVector v, SCALAR d);
  void              Set(SCALAR a, SCALAR b, SCALAR c, SCALAR d);
  void              Set(cVector normal, cVector point);
  SCALAR            Distance(cVector point);
  BOOLEAN           Intersect(cVector pointA, cVector pointB, cVector &pointS);
  cVector           Intersect(cRay ray);
};


class cMatrix {
  public:
  MATRIX            mM;
  public:
                    cMatrix();
  void              Set();
  void              Set(SCALAR a1, SCALAR a2, SCALAR a3, SCALAR a4,
                        SCALAR b1, SCALAR b2, SCALAR b3, SCALAR b4,
                        SCALAR c1, SCALAR c2, SCALAR c3, SCALAR c4,
                        SCALAR d1, SCALAR d2, SCALAR d3, SCALAR d4);
  void              Inv();
  void              MovRel(SCALAR x, SCALAR y, SCALAR z);
  void              Mul(cMatrix m);
  void              Add(cMatrix m);
  void              Scale(SCALAR sx, SCALAR sy, SCALAR sz);
  void              RotXYZ(ANGLE a, ANGLE b, ANGLE c);
  void              RotX(ANGLE a);
  void              RotY(ANGLE a);
  void              RotZ(ANGLE a);
  void              Apply(cVector &v);
  void              Apply(cVector &vOut, cVector &v);
  void              MakeProjection(cPlane plane, cVector center);
  void              MakeCameraRot(cVector pnt, cVector rot);
  void              MakeCameraPos(cVector pnt, cVector pos);
  void              MakeObjectRot(cVector pnt, cVector rot);
  void              MakeObjectPos(cVector pnt, cVector pos);
};


class cSpline {
  public:
  cVector           mV[4];
  public:
  SCALAR            GetX(SCALAR u);
  SCALAR            GetY(SCALAR u);
  SCALAR            GetZ(SCALAR u);
  cVector           Get(SCALAR u);
};


class cRay {
  public:
  cVector           mO; // origin
  cVector           mD; // direction
  public:
  void              Set(cVector pointA, cVector pointB);
};


class cSphere {
  public:
  cVector           mV;
  SCALAR            mR;
  SCALAR            mR2;  // squared radius
  public:
                    cSphere();
  void              Set(cVector *vList, INTEG n);
  BOOLEAN           Hit(cSphere other);
  BOOLEAN           Inside(cVector point);
  BOOLEAN           Hit(cRay ray, cVector &point);
};


//--[Misc Defines]------------------>


#define ppercent(zero_pc, hundred_pc, p_pc)  (zero_pc + ((hundred_pc)-(zero_pc))*(p_pc)/100 )

#define fitmin(v, lo)        ( (v) >= (lo) ? (v) : (lo) )

#define fitmax(v, hi)        ( (v) <= (hi) ? (v) : (hi) )

#define fitarea(v, lo, hi)   ( (v) <= (hi) ? (v) >= (lo) ? (v) : (lo) : (hi) )

#define inarea(v, lo, hi)    ( ( (lo) <= (v) ) && ( (v) <= (hi) ) )

#define report(m)            { if (errlog) { fprintf(errlog, (m)); fflush(errlog); } else { fprintf(stderr, (m)); fflush(stderr); }; }

#define setflag(a, b)        (a) = ( FLAGS((a)) | FLAGS((b)) )

#define resetflag(a, b)      (a) = ( FLAGS((a)) & ( ~ FLAGS((b)) ) )


#define xsqr(x)              ( (x) * (x) )

#define xsgn(v)              ((v<0)?(-1):((v>0)?(+1):(0)))

#define xabs(v)              ((v<0)?(-v):(v))

#define xmod(a, b)           ((a) -= INTEG((b)) * (INTEG( (a) ) / INTEG( (b) )))

#define xmax(v1,v2)          ((v1>v2)?(v1):(v2))

#define xmin(v1,v2)          ((v1<v2)?(v1):(v2))


//--[Parametric Spline Defines]--->


// -1.0 <= u <= 2.0

#define spline_b1(u)    ((u) * ((u) - 1) * ((u) - 2) / (-6.0F))

#define spline_b2(u)    (((u) + 1) * ((u) - 1) * ((u) - 2) / (2.0F))

#define spline_b3(u)    (((u) + 1) * (u) * ((u) - 2) / (-2.0F))

#define spline_b4(u)    (((u) + 1) * (u) * ((u) - 1) / (6.0F))

#define spline_func(a, b, c, d, u) ((a) * spline_b1((u)) + (b) * spline_b2((u)) + (c) * spline_b3((u)) + (d) * spline_b4((u)))


//--[Geometrical Formula Defines]--->


#define negVector(v)             { v[X] = -v[X]; v[Y] = -v[Y]; v[Z] = -v[Z]; }

#define setVector(v, x,  y, z)   { v[X] = x; v[Y] = y; v[Z] = z; }


//--[return value]-->


#define distance2(A, B)      ( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y])+((A)[Z]-(B)[Z])*((A)[Z]-(B)[Z]) )

#define distance(A, B)       (sqrt( ((A)[X]-(B)[X])*((A)[X]-(B)[X])+((A)[Y]-(B)[Y])*((A)[Y]-(B)[Y])+((A)[Z]-(B)[Z])*((A)[Z]-(B)[Z]) ))

#define manhatan(A, B)       ( xabs((A)[X]-(B)[X]))+xabs((A)[Y]-(B)[Y]))+xabs((A)[Z]-(B)[Z]))  )


// --Following -------------------->
// n = normal of plane;
// p = point of line/ray;
// v = slope of line/ray
// d = "D" from the plane equation


#define calcT(n, p, v, d)            (-(n[X]*p[X]+n[Y]*p[Y]+n[Z]*p[Z]+d)/(n[X]*v[X]+n[Y]*v[Y]+n[Z]*v[Z]))

#define calcP(nrml, pnt, d)          (nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)

#define calcA(nrml, pnt, d)          (-(nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)/pnt[X])

#define calcB(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Z]*pnt[Z]+d)/pnt[Y])

#define calcC(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+d)/pnt[Z])

#define calcD(nrml, pnt)             (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]))

#define calcX(nrml, pnt, d)          (-(nrml[Y]*pnt[Y]+nrml[Z]*pnt[Z]+d)/nrml[X])

#define calcY(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Z]*pnt[Z]+d)/nrml[Y])

#define calcZ(nrml, pnt, d)          (-(nrml[X]*pnt[X]+nrml[Y]*pnt[Y]+d)/nrml[Z])

#define magnitude(A)                 ( sqrt( xsqr((A)[X]) + xsqr((A)[Y]) + xsqr((A)[Z]) ) )

#define dotprod(A, B)                ((A)[X] * (B)[X] + (A)[Y] * (B)[Y] + (A)[Z] * (B)[Z] )


#define clip(a1, b1, a2, b2, clp)    ( (a1) - ( (a1) - (a2) ) / ( (b1) - (b2) ) * ( (b1) - (clp) ) )

#define clif(b1, b2, f)              ( (b1) - ( (b1) - (b2) ) * (f) )

#define clic(b1, b2, f, d)           ( (b1) - ( (b1) - (b2) ) * (f) / (d) )
  

//--[no return value]-->


#define cross(C, A, B)         { (C)[X] = (A)[Y] * (B)[Z] - (A)[Z] * (B)[Y]; (C)[Y] = (A)[Z] * (B)[X] - (A)[X] * (B)[Z]; (C)[Z] = (A)[X] * (B)[Y] - (A)[Y] * (B)[X]; }

#define vectoradd(C, A, B)     { (C)[X] = (A)[X] + (B)[X]; (C)[Y] = (A)[Y] + (B)[Y]; (C)[Z] = (A)[Z] + (B)[Z]; }

#define vectorsub(C, A, B)     { (C)[X] = (A)[X] - (B)[X]; (C)[Y] = (A)[Y] - (B)[Y]; (C)[Z] = (A)[Z] - (B)[Z]; }

#define scalarmul(C, A, B)     { (C)[X] = (A)[X] * (B); (C)[Y] = (A)[Y] * (B); (C)[Z] = (A)[Z] * (B); }

#define scalardiv(C, A, B)     { (C)[X] = (A)[X] / (B); (C)[Y] = (A)[Y] / (B); (C)[Z] = (A)[Z] / (B); }


//--[misc vec/mtx functions]-->


void normalize(VECTOR &result, VECTOR &A);

void makeNormal(VECTOR &normal, VECTOR &p0, VECTOR &p1, VECTOR &p2);

INTEG intersection(VECTOR &normal, SCALAR d,
                   VECTOR &p0, VECTOR &p1,
                   VECTOR &pnt_on_plane);

void setMatrix(MATRIX &M, SCALAR a1, SCALAR a2, SCALAR a3, SCALAR a4,
                                 SCALAR b1, SCALAR b2, SCALAR b3, SCALAR b4,
                                 SCALAR c1, SCALAR c2, SCALAR c3, SCALAR c4,
                                 SCALAR d1, SCALAR d2, SCALAR d3, SCALAR d4);

void setMatrix(MATRIX &M);

void applyMatrix(MATRIX &M, VECTOR &N);

void applyMatrix(MATRIX &M, VECTOR &N, VECTOR &outN);

void mulMatrix(MATRIX &M, MATRIX &N);

void addMatrix(MATRIX &M, MATRIX &N);

void RXMatrix(MATRIX &M, ANGLE x);

void RYMatrix(MATRIX &M, ANGLE y);

void RZMatrix(MATRIX &M, ANGLE z);

void TMatrix(MATRIX &M, SCALAR x, SCALAR y, SCALAR z);

void SMatrix(MATRIX &M, SCALAR x, SCALAR y, SCALAR z);

void RXVector(VECTOR &V, ANGLE x);

void RYVector(VECTOR &V, ANGLE y);

void RZVector(VECTOR &V, ANGLE z);


#endif
